# Copyright 2020 Cole Wilson and other contributors
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
import os
import sys
import toml
import glob
import shutil
import re
import requests
from semver import VersionInfo
from semver import compare

prefix = os.path.dirname(os.path.abspath(__file__))+os.sep+'resources'


def main(ids,arguments,nointeraction=False):
	# ============== Get Data ===============================================
	if not os.path.isfile('.'+os.sep+'sailboat.toml'):
		print('Please create a config file with `sailboat wizard` first.')
		sys.exit(0)
	try:
		data = toml.loads(open('.'+os.sep+'sailboat.toml').read())
	except toml.decoder.TomlDecodeError as e:
		print('Config error:\n\t'+str(e))
		exit()
	# ============== Get VersionInfo ===============================================
	if 'latest_build' not in data:
		data['latest_build'] = '0.0.1'
	if len(ids) >= 2: #Something provided
		if VersionInfo.isvalid(ids[1]):
			version = ids[1]
		elif ids[1].startswith('maj'):
			version = str(VersionInfo.parse(data['latest_build']).bump_major())
		elif ids[1].startswith('min'):
			version = str(VersionInfo.parse(data['latest_build']).bump_minor())
		elif ids[1].startswith('pat'):
			version = str(VersionInfo.parse(data['latest_build']).bump_patch())
		elif ids[1].startswith('pre') or ids[1].startswith('dev'):
			version = str(VersionInfo.parse(data['latest_build']).bump_prerelease())
		else:
			print('Unknown version `{}`'.format(ids[1]))
			sys.exit(0)
	else:
		try:
			latestcommit = os.popen('git rev-parse --short HEAD').read().replace('\n','')
		except KeyboardInterrupt:
			latestcommit = "build"
		if latestcommit in data['latest_build']:
			version = str(VersionInfo.parse(data['latest_build']).bump_build())
		else:
			version = str(VersionInfo.parse(data['latest_build']).replace(build=latestcommit+".1"))
	if compare(version,data['latest_build']) == -1 and not (ids[1].startswith('pre') or ids[1].startswith('dev')):
		if input(f'\u001b[31mYou are building a version ({version}) that comes before the previously built version ({data["latest_build"]}). Do you wish to continue? [y/n] \u001b[0m')[0]=='n' or nointeraction:
			print()
			sys.exit(0)
	print('\nPreparing to build version {}\n'.format(version))

	# ============== Pre-build script ===============================================
	if 'build_script' in data['build']:
		try:
			buildscript = __import__(data['build']['build_script'].replace('.py',''))
		except BaseException as e:
			print('Error with custom prebuild script:\n'+str(e))
			pass
	try:
		newdata = buildscript.pre(version,data)
		if isinstance(newdata,dict):
			data = newdata
	except BaseException as e:
		print('Error with custom prebuild script:\n\t`'+str(e)+"`")
		pass
	# ============== Show what will happen ===============================================
	print('This command will build the following:')
	print('\t- Generate a correct directory structure, setup.py, .gitignore, etc...')
	only=False
	for x in arguments:
		if '-only' in x[0]:
			only=True
			break
	dopypi,dobrew,domac,dowin,doact,doset=False,False,False,False,False,False
	if only:
		if ('--pypi-only','') in arguments:
			dopypi = True
		elif ('--homebrew-only','') in arguments:
			dobrew = True
		elif ('--windows-only','') in arguments and sys.platform.startswith('win'):
			dowin = True
		elif ('--mac-only','') in arguments and sys.platform.startswith('darwin'):
			domac = True
		elif ('--actions-only','') in arguments:
			doact = True
		elif ('--setup-only','') in arguments:
			doset = True
		elif ('--unix-only','') in arguments:
			donix = True
		
	elif 'actions_only' in data['build'] and data['build']['actions_only'] and not ('CI' in os.environ and os.environ['CI']=="TRUE"):
		dopypi = True
		dobrew = True
		dowin = False
		domac = False
		donix = False
		doact = True
		doset = True
	else:
		dopypi = True
		dobrew = data['build']['homebrew']
		dowin = data['build']['windows'] and sys.platform.startswith('win')
		domac = data['build']['mac'] and sys.platform.startswith('darwin')
		donix = data['build']['unix'] and sys.platform.startswith('l')
		doact = data['build']['actions']
		doset =True
	if ('--no-installer','') in arguments:
		doinstall=False
	else:
		doinstall=data['build']['installer']
	if dopypi:
		print('\t- A distributable Python module.')
	if dobrew:
		print('\t- A Homebrew formula.')
	if dowin:
		installer = " with a .msi installer." if doinstall else "."
		print('\t- A Windows app'+installer)
	if domac:
		installer = " with a .dmg installer." if doinstall else "."
		print('\t- A Mac app'+installer)
	if donix:
		print('\t- A Unix executable')
	
	if doact:
		print('\t- A GitHub Actions file for building Mac and Windows apps and publishing a Github release.')
	if not nointeraction:
		input('Press enter to continue.')
	# ============== Create bin Script ===============================================
	try:
		os.mkdir('bin')
	except:
		pass
	bins = []
	for commandname in data['build']['commands'].keys():
		if data['build']['commands'][commandname] == '':
			open("bin"+os.sep+commandname,'w+').write(f"#!"+os.sep+"usr"+os.sep+"bin"+os.sep+f"env bash\npython3 -m {data['short_name']} $@")
			bins.append('bin'+os.sep+commandname)
	# ============== Get module names ===============================================
	if 'no_import' not in data['resources']:
		data['resources']['no_import'] = []
	mods = []
	for x in glob.glob(data['short_name']+os.sep+'*.py'):
		f = open(x)
		mods += re.findall('(?m)(?:from[ ]+(\S+)[ ]+)?import[ ]+\S+?:[ ]+as[ ]+\S+?[ ]*$',f.read())
		f.close()
	modules = []
	for x in mods:
		modules.append(x[1].split('.')[0])
	for module in set(modules):
		if module not in data['resources']['no_import'] and ( module!= data['short_name'] and module not in sys.builtin_module_names and module not in data['resources']['modules']):
			print('Checking for {} on PyPi...'.format(module))
			response = requests.get("https://pypi.python.org/pypi/{}/json".format(module))
			if response.status_code == 200:
				data['resources']['modules'].append(module)
			else:
				data['resources']['no_import'].append(module)

	# ============== Generate setup.py ===============================================
	print('\n\n\u001b[4m\u001b[1;36mGenerating setup.py\u001b[0m')

	if doset:
		if 'custom_setup' in data:
			cu=str(data['custom_setup'])
		else:
			cu=str({})
		with open(prefix+os.sep+'setup.py.template') as datafile:
			template = datafile.read()

		entries = []
		for commandname in data['build']['commands'].keys():
			if data['build']['commands'][commandname]!="":
				modname = ".".join(data['build']['commands'][commandname].split('.')[:-1])
				funcname = data['build']['commands'][commandname].split('.')[-1]
				entries.append(commandname+"="+data["short_name"]+"."+modname+":"+funcname)
		try:
			pyv = version.split('+')[0]
		except:
			pyv = version

		setup = template.format(
			**data,
			**data['resources'],
			cu=cu,
			bins=bins,
			version = pyv,
			entry_points = entries
			
		)
		open('setup.py','w+').write(setup)

	# ============== Generate directory structure ===============================================
	print('\n\n\u001b[4m\u001b[1;36mGenerating Directory Structure\u001b[0m')

	if not os.path.isfile('.gitignore'):
		open('.'+os.sep+'.gitignore','w+').write(open(prefix+os.sep+'gitignore.template').read().replace('/',os.sep))
	source_dir = os.getcwd()
	target_dir = data["short_name"]+os.sep
	types = ('*.py',*data['resources']["data_files"])
	file_names = []
	for files in types:
		file_names.extend(glob.glob(files))
	if not os.path.isdir(target_dir):
		os.mkdir(target_dir)
	try:
		bs = data['build']['build_script']
	except:
		bs = "RANDOM-----edfskjsdhflkjdhflksdjhflkasjdhflkasjdhflkasjdhflkajsdhflkjadshf"
	for file_name in file_names:
		if file_name in ("setup.py","sailboat.toml",bs):
			continue
		shutil.move(os.path.join(source_dir, file_name), target_dir+os.sep+file_name)
	for filename in glob.glob(target_dir+os.sep+'LICE*'):
		shutil.copyfile(filename,'LICENSE')
	if not os.path.isfile(target_dir+'__init__.py'):
		open(target_dir+'__init__.py','w+').write('# This file must exist, empty or not')
	if data['resources']['file']!="" and not os.path.isfile(data['short_name']+os.sep+'__main__.py'):
		try:
			os.rename(data['short_name']+os.sep+data['resources']['file'],data['short_name']+os.sep+'__main__.py')
			open(data['short_name']+os.sep+data['resources']['file'],'w+').write('# Please edit __main__.py for the main code. Thanks!\n(you can delete this file.)')
		except FileNotFoundError:
			pass
	
	# ============== Generate pypi files ===============================================
	print('\n\n\u001b[4m\u001b[1;36mGenerating PyPi files...\u001b[0m')

	if dopypi:
		try:
			shutil.rmtree('dist')
		except:
			pass
		os.system('python3 .'+os.sep+'setup.py bdist_wheel sdist -d dist'+os.sep+'pypi')
		try:
			shutil.rmtree('build')
		except:
			pass
		for x in glob.glob('dist'+os.sep+'*.whl'):
			os.rename(x,x.replace('dist'+os.sep,'dist'+os.sep+'pypi'+os.sep))
		for x in glob.glob('*.egg-info'):
			shutil.rmtree(x)

	# ============== Generate homebrew file ===============================================
	print('\n\n\u001b[4m\u001b[1;36mGenerating Homebrew file...\u001b[0m')

	if dobrew:
		retmp = '   resource "{name}" do\n      url "{url}"\n      sha256 "{sha256}"\n   end\n'
		resources = ''
		for modulename in data['resources']['modules']:
			req = requests.get('https://pypi.org/pypi/{}/json'.format(modulename)).json()
			versionPy = req['info']['version']
			url = req['releases'][versionPy][0]['url']
			sha256 = req['releases'][versionPy][0]['digests']['sha256']
			if not (url.endswith('.tar.gz') or url.endswith('.zip')):
				try:
					url = req['releases'][versionPy][1]['url']
					sha256 = req['releases'][versionPy][1]['digests']['sha256']
				except:
					continue
			resources+=retmp.format(name=modulename,url=url,sha256=sha256)
		os.makedirs('dist'+os.sep+'homebrew')
		f = open('dist'+os.sep+'homebrew'+os.sep+'{name}.rb'.format(name=data['name']),'w+')
		f.write(open(prefix+os.sep+'brew.rb').read().format(
			**data,
			resources2 = resources,
			version = version
		))
		f.close()
	# ============== Generate w/Pyinstaller ===============================================
	print('\n\n\u001b[4m\u001b[1;36mGenerating Pyinstaller files...\u001b[0m')

	# domac = True
	if dowin or domac or donix:
		try:
			import PyInstaller.__main__
		except:
			print('Installing PyInstaller...')
			os.system('pip3 install pyinstaller')
			import PyInstaller.__main__
		
		mods = []
		for x in data['resources']['modules']:
			mods.append('--hidden-import')
			mods.append(x)

		datafiles = []
		for x in data['resources']['data_files']:
			for g in glob.glob(data['short_name']+os.sep+x):
				datafiles.append('--add-data')
				datafiles.append(g+os.pathsep+g.replace(data['short_name']+os.sep,''))
		typ = '--nowindowed' if data['build']['type']=='1' else '--noconsole'

		ico = ['--icon',data['resources']['icon']] if 'icon' in data['resources'] else []
		options = [
			data['short_name']+os.sep+'__main__.py',
			'--onefile',
			'--name',
			data['name'] if ('-n','')not in arguments else 'app',
			'--distpath',
			'.'+os.sep+'dist'+os.sep+'pyinstaller',
			*mods,
			*datafiles,
			typ,
			*ico,
			'--osx-bundle-identifier',
			data['build']['bundle_id']
		]
		print(options)
		PyInstaller.__main__.run(options)
		try:
			print('removing '+data['name']+".spec...")
			os.remove(data['name']+".spec")
		except:
			print('removing app.spec...')
			os.remove("app.spec")			
	# ============== Mac .app Bundle ===============================================
	print('\n\n\u001b[4m\u001b[1;36mGenerating Mac .app bundle...\u001b[0m')
	if domac:
		os.chdir('dist')

		os.mkdir(data['name'])
		os.chdir(data['name'])

		os.mkdir('Contents')
		os.chdir('Contents')

		os.mkdir('MacOS')
		os.mkdir('Resources')

		infoPlist = open('Info.plist','w+')
		infoPlist.write(open(prefix+'/info.plist.xml').read().format(
			**data,
			**data['build'],
			version = version
		))
		infoPlist.close()

		shutil.copy('./../../pyinstaller/'+data['name'],'MacOS')
		os.chdir('./../../..')

		os.rename('./dist/'+data['name'],'./dist/'+data['name']+".app")
	else:
		print('not generating mac .app bundle because on {} not mac.'.format(sys.platform))
	# ============== Generate Installer Package ===============================================
	print('\n\n\u001b[4m\u001b[1;36mGenerating Installer Package...\u001b[0m')

	if sys.platform.startswith('win') and doinstall and False:#WINDOWS
		os.system('pip install distro')
		os.system('pip install git+https://github.com/x24git/wixpy')
		d = open(prefix+os.sep+'wixpy.template.json').read().format(
			**data,
			version=version,
			icns=data['resources']['icon'],
			keywo=", ".join(data['keywords'])
		)
		open('wixpy.json','w+').write(d)
		print(d)
		print('STARTING WIXPY...')
		os.system('wix.py wixpy.json')

	elif sys.platform.startswith('darwin'):#MAC
		try:
			import dmgbuild
		except:
			print('Installing dmgbuild')
			os.system('pip3 install dmgbuild')
			import dmgbuild
		del dmgbuild
		if not os.path.isdir('build'):
			os.mkdir('build')
		open('build'+os.sep+'settings.py','w+').write(open(prefix+os.sep+'settings.py.template').read().format(
			**data,
			version=version,
			icns=data['resources']['icon'],
			keywo=", ".join(data['keywords'])
		))
		os.system(f'cat build/settings.py;dmgbuild -s .{os.sep}build{os.sep}settings.py "{data["name"]} Installer" ./{data["name"]}.dmg')

	else:
		print(f'Installer creation not yet supported for {sys.platform}!')
	# ============== Generate Github Actions Workflow ===============================================
	print('\n\n\u001b[4m\u001b[1;36mGenerating GitHub Actions File...\u001b[0m')
	if doact:
		try:
			oldact = open('.github'+os.sep+'workflows'+os.sep+'sailboat.yml').read().split('\n')[0]
		except:
			oldact="\n"
		try:
			f = open('.github'+os.sep+'workflows'+os.sep+'sailboat.yml','w+')
		except:
			os.system('mkdir -p .github'+os.sep+'workflows'+os.sep)
			f = open('.github'+os.sep+'workflows'+os.sep+'sailboat.yml','w+')
		newdata = open(prefix+os.sep+'sailboat.yml.template').read().format(
			**data,
			mac=""if data['build']['mac']else"#",
			windows=""if data['build']['windows']else"#",
			win_ext=".exe"if data['build']['installer']else".exe",
			mac_ext=".dmg"if not data['build']['unix']else"",
			u=""if data['build']['unix']else"#"
		).replace('\t','  ')
		f.write(newdata)

		f.close()
		newact = open('.github'+os.sep+'workflows'+os.sep+'sailboat.yml').read().split('\n')[0]
		data['build']['actions_built_latest'] = newact != oldact
		print(oldact)
		print(newact)
	# ============== Post build ===============================================
	try:
		newdata = buildscript.post(version,data)
		if isinstance(newdata,dict):
			data = newdata
	except:
		pass
	# ============== Save VersionInfo ===============================================
	data['latest_build'] = version
	open('sailboat.toml','w+').write(toml.dumps(data))
	os.system('python .'+os.sep+'setup.py develop')


from sailboat.plugins import Plugin
import pkg_resources
import os
import glob
import sys
import json
from pathlib import Path
from semver import VersionInfo


import traceback

def refreshEntries():
	plugins={"core":{},"build":{},"release":{},"command":{}}
	for entry_point in pkg_resources.iter_entry_points(group='sailboat_plugins'):
		temp = entry_point.load()
		plugins[temp._type][entry_point.name] = {
			"show" : temp._show,
			"dist" : str(entry_point.dist).split(' ')[0],
			"description" : temp.description,
			"type" : temp._type,
			"release" : temp._release,
			"order" : temp._order,
			"default_os": str(temp._os)
		}
	try:
		path = Path(__file__)
		prefix = os.wwpath.abspath(path.parent)
	except:
		prefix = os.path.dirname(os.path.abspath(__file__))+os.sep
	f = open(prefix+os.sep+'plugins.json','w+')
	f.write(json.dumps(plugins, indent=2, sort_keys=True))
	f.close()

	del f

class QuickStart(Plugin):
	_type = "core"
	description = "Get your project up and running."

	def runPlugin(self,plug,plugins,opts=[]):
		b = pkg_resources.load_entry_point('sailboat','sailboat_plugins',plug)
		temp = b(
			data=self.data,
			options=opts,
			name=plug,
			prefix=self.prefix,
			version=None
		)
		temp.run(plugins = plugins)
		return temp.data

	def run(self,plugins={},**kwargs):
		input('This quickstart command will get you started on your project.\nFirst, it will set up the config file and your GitHub settings. Then it will add suggested plugins for your project!\n\nPress enter to continue, and ctrl+c to skip a step.')
		for plug in ['wizard', 'git']:
			print("\n\n"+self.red(plug)+"\n"+(len(plug)*"-")+"\n")
			self.data = self.runPlugin(plug,plugins)
		print()
		print("\n\n"+self.red('plugins')+"\n"+(len('plugins')*"-")+"\n")
		pypi = "pypi" if input('Would you like to generate pypi files for your project? [Y/n]')+"y"[0] == 'y' else ""
		pyinstaller = "pyinstaller" if input('Would you like to generate pyinstaller files for your project? [Y/n]')+"y"[0] == 'y' else ""
		homebrew = "homebrew" if input('Would you like to generate homebrew files for your project? [Y/n]')+"y"[0] == 'y' else ""
		combined = f"{pypi} {pyinstaller} {homebrew}"
		if len(combined) > 0:
			print()
			self.runPlugin('add',plugins,opts=[*combined.split()])

		print('Your project is set up, run `sail build` to build, or `sail add <plugin>` to add a new plugin.')

class ManagePlugins(Plugin):
	_type = "core"
	description = "plugin manager."
	def autocomplete(self):
		return ['a','b']
	def run(self,plugins={},**kwargs):
		if self.options == []:
			print("usage: sail plugins [refresh]\n\n\trefresh: reload previously installed plugins.")
		elif self.options == ['refresh'] or self.options == ['-r']:
			refreshEntries();
			print('Done!')
		elif self.options == ['list']:
			print('Build:')
			for x in self.data['build']:
				print('\t- '+x)
			print('Release:')
			for x in self.data['release']:
				print('\t- '+x)
			print('Command:')
			for x in self.data['command']:
				print('\t- '+x)
			
		else:
			print('sailboat: error: invalid option `{}`.'.format(self.options[0]))

class Develop(Plugin):
	_type = "core"
	description = "run your project without building it."
	def run(self,plugins={},**kwargs):
		if os.path.isfile('setup.py'):
			os.system('python setup.py develop')
		else:
			self.info("Can't find setup.py. Create one using sail build.")
class Wizard(Plugin):
	_type = "core"
	description = "configure you project or a plugin."
	setup = { # Top Level Setup
		"name::str":"Full name of your project: ",
		"short_name::str":"Short name of your project: ",
		"email::str":"Email for your project: ",
		"author::str":"Your name(s): ",	
		"short_description::str":"Short description of your project: ",
		"description::str":"Long description of your project: ",
		"url::str":"Home URL of your project",
		"keywords::str":"Your project's keywords, seperated with a a space.",
	}

	def wizard(self):
		super().wizard()
		lt = "BSD-2-Clause/BSD-3-Clause/Apache-2.0/LGPL-2.0/LGPL-2.1/LGPL-3.0/GPL-2.0/GPL-3.0/CCDL-1.0/GNU LGPL/MIT/Other SPDX License ID"
		licen = len(glob.glob('.'+os.sep+'LICENS*'))>0
		if licen:
			if 'license' not in self.data:
				print('\033[1;34mYou seem to have a license file in your project, but what type is it?')
				for license in lt.split(os.sep+''):
					print(f'\t- {license}')
				self.data['license'] = input(">>>\033[0m ")
			else:
				print('\u001b[34mLicense:\u001b[0m {}'.format(self.data['license']))
		else:
			self.data['license'] = "none"
		print(self.section('Resource Settings:'))
		if 'resources' not in self.data:
			self.data['resources']={}
		questions = {
			"icon":"Square .png icon for your project, leave blank if none",
			"data_files":"List of all data files for your project seperated with spaces. * counts as wildcard",
			"modules":"Space seperated list of modules required for your project",
			"file":"Main python file of your project. Leave blank if none"
		}
		for key in questions:
			if key in self.data['resources']:
				print(self.blue(questions[key])+': '+str(self.data['resources'][key]))
			else:
				if key in ('modules','data_files'):
					self.data['resources'][key] = input(self.blue(questions[key])+": ").split()
				else:
					self.data['resources'][key] = input(self.blue(questions[key])+": ")
		self.data['name'] = self.data['name'].replace(' ','_')
		self.data['short_name'] = self.data['short_name'].replace(' ','_')
		
		print(self.red('Be sure to prefix any paths to any resources with \u001b[4m`os.path.dirname(os.path.abspath(__file__))+os.sep`\u001b[0m\033[1;31m to make sure that they use the correct path and not the current directory.'))
	def run(self,plugins={}):
		if self.options == []:
			self.wizard()
			for tt in ['command','build','release']:
				for x in self.data[tt]:
					print(self.section(x.title()+":"))
					dist = plugins[tt][x]['dist']
					b = pkg_resources.load_entry_point(dist,'sailboat_plugins',x)
					b = b(
							data=self.data,
							options=[],
						name=x,
						prefix=self.prefix,
							version=None
					)
					try:
						b.wizard()
					except KeyboardInterrupt:
						print('Abborted by user')
					except BaseException:
						print(traceback.print_exception(*sys.exc_info()))
						print('Error in `{}` wizard script:\n\t{}'.format(x,traceback.print_exception(*sys.exc_info())))
					self.data[tt][x] = b.data[tt][x]
			
		else:
			for x in self.options:
				if x in plugins['command']:
					t = 'command'
				elif x in plugins['release']:
					t = 'release'
				elif x in plugins['build']:
					t = 'build'
				else:
					print("Couldn't find the {} plugin.".format(x))
					continue
				dist = plugins[t][x]['dist']
				temp = pkg_resources.load_entry_point(dist,'sailboat_plugins',x)
				temp = temp(
					data=self.data,
					options=[],
					name=x,
					prefix=self.prefix,
					version=None
				)
				try:
					temp.wizard()
				except KeyboardInterrupt:
					print('Abborted by user')
				except BaseException as e:
					print('Error in `{}` wizard script:\n\t{}'.format(x,e))
				self.data[t][x] = temp.data[t][x]

class Git(Plugin):
	_type = "core"
	description = "Manage Git for your project."

	def run(self,plugins={},**kwargs):
		if self.options == ['push']:
			a = input("Message: ").replace('"',r'"')
			os.system(f'git add .;git commit -a -m "{a}";git push;')
			return
		if 'github' not in self.data['git']:
			uname = input('GitHub username: ')
			if input('Do you have a GitHub repository for this project yet?')[0]=='y':
				self.data['git']['github'] = uname+'/'+input('GitHub repo name: ')
			else:
				print('Go to https://github.com/new and create a repo.')
				self.data['git']['github'] = uname+'/'+input('GitHub repo name: ')
			os.system(f"""git init;
git add .;
git config --global credential.helper "cache --timeout=3600";
git config user.name "{self.data["author"]}";
git config user.email "{self.data["email"]}";
git commit -m "Initial Commit :rocket:";
git remote add origin https://github.com/{self.data['git']['github']}.git;
git push -u origin master;
""")
		print('Your GitHub repo is setup. To update your repo, type:\n\n\tgit add .;git commit -a -m "Your message here";git push;\n\nin your terminal, or `sail git push`')

class GithubRelease(Plugin):
	_type = "release"
	description = "GitHub Tagged Release"
	def release(self):
		os.system(f"""git init;
git add .;
git config --global credential.helper "cache --timeout=3600";
git config user.name "{self.data["author"]}";
git config user.email "{self.data["email"]}";
git commit -m "{self.data['release-notes']}";
git tag v{self.version};
git remote add origin https://github.com/{self.data['git']['github']}.git||echo origin already added;
git push -u origin master --tags;
""")

class Release(Plugin):
	_type = "core"
	description = "Release your project."
	def run(self,plugins={},**kwargs):
		self.data['release-notes'] = input('Release Notes: ')
		version = VersionInfo.parse(self.data['latest_build'])
		version = str(VersionInfo(major=version.major,minor=version.minor,patch=version.patch,prerelease=version.prerelease))
		runs = {}
		if self.options == []:
			for x in self.data['release']:
				if x in plugins['release']:
					runs[x] = plugins['release'][x]['order']
			for x in self.data['build']:
				if x in plugins['build']:
					if plugins['build'][x]['release']:
						runs[x] = plugins['build'][x]['order']
			for x in self.data:
				if x in plugins['core']:
					if plugins['core'][x]['release']:
						runs[x] = plugins['core'][x]['order']
			
		else:
			for x in self.options:
				if x in self.data['release'] and x in plugins['release']:
					runs[x] = plugins['release'][x]['order']
				elif x in self.data['build']:
					if plugins['build'][x]['release']:
						runs[x] = plugins['build'][x]['order']
				elif x in self.data:
					if plugins['core'][x]['release']:
						runs[x] = plugins['core'][x]['order']
				
				else:
					print('sailboat: error: {} is not a valid release plugin.'.format(x))
					return
		runstemp = []
		for x in runs:
			runstemp.append(f"{runs[x]}::{x}")
		runstemp.sort()
		runs = []
		for x in runstemp:
			order,name = x.split('::')
			runs.append(name)
		input(f'Press enter to release version {version} the following ways:\n\t- '+'\n\t- '.join(runs)+'\n\n>>>')
		dones = []
		lorder = 0
		for release_plugin in runs:
			if release_plugin in dones:
				continue
			print(self.section(release_plugin+":"))
			if release_plugin in self.data['release']:
				dist = plugins['release'][release_plugin]['dist']
			elif release_plugin in self.data['build']:
				dist = plugins['build'][release_plugin]['dist']
			else:
				dist = plugins['core'][release_plugin]['dist']
			temp = pkg_resources.load_entry_point(dist,'sailboat_plugins',release_plugin)
			temp = temp(
				data=self.data,
				options=[self.options],
				name=release_plugin,
				prefix = self.prefix,
				version=version
			)
			temp.release()
			if temp._type == 'core':
				self.dat = temp.data
			else:
				self.data[temp._type][release_plugin] = temp.data[temp._type][release_plugin]
			dones.append(release_plugin)
			lorder = order
		print()
		self.data['latest_release'] = version

		


class Remove(Plugin):
	_type = "core"
	description = "remove a plugin from you project."
	def run(self,plugins,**kwargs):
		if len(self.options)<1:
			print('usage: sail remove [plugin ...]\n\n\tRemoves plugins from project.')
			return
		for option in self.options:
			t = ''
			if option in plugins['command'].keys():
				t = 'command'
			if option in plugins['release'].keys():
				t = 'release'
			if option in plugins['build'].keys():
				t = 'build'
			if t=='':
				print('couldn`t find {}, therfore not removed.'.format(option))
			else:
				try:
					del self.data[t][option]
					print('removed {}'.format(option))
				except KeyError:
					print('couldn`t find {}, therfore not removed.'.format(option))

class Add(Plugin):
	_type = "core"
	description = "add a plugin to your project."
	def run(self,**kwargs):
		# get all availbile plugin
		done = []
		if len(self.options)==0:
			print('''usage: sail add [plugins ...]

	Install PLUGINS to current project.
	''')
		for point in pkg_resources.iter_entry_points('sailboat_plugins'):
			if (point.name in self.data['build'] or point.name in self.data['command'] or point.name in self.data['release']) and point.name in self.options:
				print('{} already added'.format(point.name))
				continue
			if point.name in self.options:
				done.append(point.name)
				temp = point.load()
				run = temp(data=self.data,options=[],name=point.name,prefix=self.prefix)
				print(self.section(f'{point.name} [{run._type}]:{run.description}.'))
				if point.name not in self.data[run._type]:
					self.data[run._type][point.name] = {}
				if run._type == 'core':
					print('BE WARNED: THIS PLUGIN TYPE IS REGISTERED AS "core", which means it has unfiltered access to your project`s data. Please be 100% sure you want to do this.')
				run.data = self.data
				try:
					run.wizard()
					run.add()
				except KeyboardInterrupt:
					print('Aborted by user')
				except BaseException as e:
					print('Error in `{}` wizard script:\n\t{}'.format(point.name,e))
				data = run.data
				self.data[run._type][point.name] = data[run._type][point.name]
				print('added {}'.format(point.name))
		for x in self.options:
			if x not in done and (x not in self.data['build'] and x not in self.data['command'] and x not in self.data['release']):
				print('error: could not find {}, please make sure you have downloaded it.'.format(x))
		refreshEntries()
		return self.data

class Actions(Plugin):
	description = "Generate a GH Actions workflow file."
	_type = "core"
	
	def run(self,plugins,**kwargs):
		linux = ""
		mac = ""
		windows = ""
		for x in self.data['build']:
			if 'windows' in plugins['build'][x]['default_os']:
				windows = windows + " " + x
			if 'linux' in plugins['build'][x]['default_os']:
				linux = linux + " " + x
			if 'mac' in plugins['build'][x]['default_os']:
				mac = mac + " " + x

		with self.getResource(f'resources{os.sep}sailboat.yml.template') as temp:
			new = temp.read().format(
				linux=linux,
				mac=mac,
				windows=windows,
				l="#" if linux == "" else "",
				m="#" if mac == "" else "",
				w="#" if windows == "" else "",
				**self.data
			)
			try:
				f = open(f'.github{os.sep}workflows{os.sep}sailboat.yml','w+')
			except:
				os.mkdir('.github')
				os.mkdir(f'.github{os.sep}workflows')
				f = open(f'.github{os.sep}workflows{os.sep}sailboat.yml','w+')
			f.write(new)
			f.close()
		
		print('Workflow generated. Remember to push twice in order for new workflow to take effect.')


from sailboat.plugins import Plugin

import os
import glob
import shutil
import re
import requests
import sys
import traceback
import pkg_resources

from semver import VersionInfo
from semver import compare
		
class Build(Plugin):
	_type = "core"
	description = "build your project for release."
	def run(self,plugins,**kwargs):
		if len(self.options)>0 and self.options[0] == 'help':
			print("usage: sail build [version (increment)]\n\tThis command builds your project using the sailboat.toml file.\n\tValid options for version:\n\t\t- Valid semver.org string: set that as version\n\t\t- `major`: increments the major version by one.\n\t\t- `minor`: increments the minor version by one.\n\t\t- `patch`: increments the patch version by one.\n\t\t- `pre`: increments the prerelease version by one.\n\t\t- None: increments build version by one.")
			return
  #=====================================================================================
		if 'latest_build' not in self.data:
			self.data['latest_build'] = '0.0.1'
		if len(self.options) >= 1: #Something provided
			if VersionInfo.isvalid(self.options[0]):
				version = self.options[0]
			elif self.options[0].startswith('maj'):
				version = str(VersionInfo.parse(self.data['latest_build']).bump_major())
			elif self.options[0].startswith('min'):
				version = str(VersionInfo.parse(self.data['latest_build']).bump_minor())
			elif self.options[0].startswith('pat'):
				version = str(VersionInfo.parse(self.data['latest_build']).bump_patch())
			elif self.options[0].startswith('pre') or self.options[0].startswith('dev'):
				version = str(VersionInfo.parse(self.data['latest_build']).bump_prerelease())
			elif self.options[0].startswith('+') or self.options[0].startswith('build'):
				version = str(VersionInfo.parse(self.data['latest_build']).bump_build())
			else:
				print('Unknown version `{}`'.format(self.options[0]))
				return
		else:
			try:
				latestcommit = os.popen('git rev-parse --short HEAD').read().replace('\n','')
			except KeyboardInterrupt:
				latestcommit = "build"
			if latestcommit in self.data['latest_build']:
				version = str(VersionInfo.parse(self.data['latest_build']).bump_build())
			else:
				version = str(VersionInfo.parse(self.data['latest_build']).replace(build=latestcommit+".1"))
		if compare(version,self.data['latest_build']) == -1 and not (self.options[0].startswith('pre') or self.options[0].startswith('dev')):
			if input(f'\u001b[31mYou are building a version ({version}) that comes before the previously built version ({self.data["latest_build"]}). Do you wish to continue? [y/n] \u001b[0m')[0]=='n' or ('-y' in self.options or '--no-interaction' in self.options):
				print()
				return
		print('\nPreparing to build version {}\n'.format(version))
		self.data['latest_build'] = version
 #=====================================================================================
		print('\n\n\u001b[4m\u001b[1;36mGenerating Directory Structure\u001b[0m')
		print('This step will generate a correct directory structure for your project.')
		if not os.path.isfile('.gitignore'):
			open('.'+os.sep+'.gitignore','w+').write(self.getResource('resources'+os.sep+'gitignore.template').read().replace('/',os.sep))
		source_dir = os.getcwd()
		target_dir = self.data["short_name"]+os.sep
		types = ('*.py',*self.data['resources']["data_files"])
		file_names = []
		for files in types:
			file_names.extend(glob.glob(files))
		if not os.path.isdir(target_dir):
			os.mkdir(target_dir)
		for file_name in file_names:
			if file_name in ("setup.py","sailboat.toml",self.data['resources']['file']):
				continue
			shutil.move(os.path.join(source_dir, file_name), target_dir+os.sep+file_name)
		for filename in glob.glob(target_dir+os.sep+'LICE*'):
			shutil.copyfile(filename,'LICENSE')
		if not os.path.isfile(target_dir+'__init__.py'):
			open(target_dir+'__init__.py','w+').write('# This file must exist, empty or not')
		if self.data['resources']['file']!="" and not os.path.isfile(self.data['short_name']+os.sep+'__main__.py'):
			try:
				os.rename(self.data['short_name']+os.sep+self.data['resources']['file'],self.data['short_name']+os.sep+'__main__.py')
				open(self.data['short_name']+os.sep+self.data['resources']['file'],'w+').write('# Please edit __main__.py for the main code. Thanks!\n(you can delete this file.)')
			except FileNotFoundError:
				pass
 #=====================================================================================
		print('\n\n\u001b[4m\u001b[1;36mFetching Modules:\u001b[0m')
		print('Scanning module imports...')
		if 'no_import' not in self.data['resources']:
			self.data['resources']['no_import'] = []
		mods = []
		for x in glob.glob(self.data['short_name']+os.sep+'*.py'):
			f = open(x)
			b = f.read()
			f.close()
			mods += re.findall('^import[ ]+(.*)',b,re.M)
			mods += re.findall('^from[ ]+(.*) import',b,re.M)
		modules = []
		for x in set(mods):
			modules.append(x.split('.')[0])
		for module in set(modules):
			if module not in self.data['resources']['no_import'] and ( module!= self.data['short_name'] and module not in sys.builtin_module_names and module not in self.data['resources']['modules']):
				print('Checking for {} on PyPi...'.format(module))
				response = requests.get("https://pypi.python.org/pypi/{}/json".format(module))
				if response.status_code == 200:
					self.data['resources']['modules'].append(module)
				else:
					self.data['resources']['no_import'].append(module)
  #=====================================================================================
		try:
			shutil.rmtree('dist')
		except FileNotFoundError:
			pass
		dones = []
		if len(self.options[1:]) == 0:
			notdones = [*self.data['build'].keys()]
		else:
			notdones = self.options[1:]
		for build_plugin in notdones:
			if build_plugin not in self.options and ('_run' in self.data['build'][build_plugin] and not self.data['build'][build_plugin]['_run']):
				continue
			if build_plugin in dones:
				continue
			if build_plugin not in self.data['build'].keys():
				continue
			elif '_needs' in self.data['build'][build_plugin]:
				if isinstance(self.data['build'][build_plugin]['_needs'],str):
					self.data['build'][build_plugin]['_needs'] = [self.data['build'][build_plugin]['_needs']]
				for x in self.data['build'][build_plugin]['_needs']:
					if x not in dones:
						notdones.append(build_plugin)
						build_plugin = x
			print('\n\u001b[4m\u001b[1;36m{}:\u001b[0m'.format(build_plugin.title()))
			try:
				dist = plugins['build'][build_plugin]['dist']
				job = pkg_resources.load_entry_point(dist,'sailboat_plugins',build_plugin)
				job = job(
					data=self.data,
					options=[],
					name=build_plugin,
					prefix=self.prefix,
					version=version
				)
			except KeyError as plugin:
				sys.exit(f'You seem to have added the {plugin} plugin, but it does not appear to be installed!')
			try:
				job.run()
			except KeyboardInterrupt:
				print('\n\nUser has aborted at step {}.\n\n'.format(build_plugin))
				return
			except BaseException as error:
				print('\n\nError at step {}:\n\n\t{}\n\n'.format(build_plugin,self.red(error)))
				return
			self.data[job._type][build_plugin] = job.data[job._type][build_plugin]
			dones.append(build_plugin)
		print(self.section('Finishing up...'))
		os.system('python setup.py develop')


# Copyright 2020 Cole Wilson and other contributors
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
import os
import sys
import toml
import shutil
import requests
import time

def main(arguments,ids):
	if not os.path.isfile('.'+os.sep+'sailboat.toml'):
		print('Please create a config file with `sailboat wizard` first.')
		sys.exit(0)
	try:
		data = toml.loads(open('.'+os.sep+'sailboat.toml').read())
	except toml.decoder.TomlDecodeError as e:
		print('Config error:\n\t'+str(e))
		exit()
	if ('-f','') not in arguments:
		try:
			latestag = os.popen('git tag').read().split('\n')[-2]
		except:
			latestag = ""
		CHANGELOG = os.popen('git log {}..HEAD --oneline'.format(latestag)).read()
		chlg = "\n## CHANGELOG:\n"
		for x in CHANGELOG.split('\n')[:-1]:
			chlg+=f"[`{x[:7]}`](https://github.com/{data['git']['github']}/commits/{x[:7]}){x[7:]}\n"

		data['build']['release_notes'] = ''
		data['build']['release_notes'] +=input('One line release message:\n> ')
		print('Extended desciption: (ctrl+c to finish):')

		while True:
			try:
				data['build']['release_notes'] +='\n'+input('> ')
			except KeyboardInterrupt:
				print('\n\n')
				break
		data['build']['release_notes']+=chlg if len(chlg.split('\n'))>4 else ''
		f = open('.release-notes-latest','w+')
		f.write(data['build']['release_notes'])
		f.close()
		data['build']['release_notes'] = data['build']['release_notes'].replace('\n',' / ').replace('"','`')[3:]

	try:
		version = data['latest_build'] = data['latest_build'].split('+')[0]
	except:
		print('You must run `sailboat build` first.')
		exit()
	if "pypi" in ids or len(ids)==1:
		print("Please make sure that you have a https://pypi.org/ account.")
		try:
			import twine
		except:
			input('Press enter to continue installing `twine`. Press ctrl+c to exit.')
			os.system('python3 -m pip install --user --upgrade twine || python3 -m pip install --upgrade twine')
			import twine
		del twine
		print('\u001b[4m\u001b[1;36mPyPi Credentials:\u001b[0m')
		os.system('python3 -m twine upload dist'+os.sep+'pypi'+os.sep+'*')
		print('waiting...')
	if "brew" in ids or len(ids)==1:
		if "pypi" in runs:
			print('Must wait 20 seconds for pypi files to upload before creting homebrew formula...\n')
			for i in range(0, 20):
					time.sleep(1)
					sys.stdout.write(u"\u001b[1000D" + str(20-(i + 1)).zfill(2) + " seconds left.\t")
					sys.stdout.flush()
			print()
		try:
			shutil.rmtree('homebrew-taps')
		except:
			pass
		if data['build']['homebrew']:
			f = open('dist'+os.sep+'homebrew'+os.sep+data['name']+'.rb')
			oldFormula = f.read()
			f.close()
			f = open('dist'+os.sep+'homebrew'+os.sep+data['name']+'.rb','w+')
			req = requests.get('https://pypi.org/pypi/{}/json'.format(data['short_name'])).json()
			versionPy = req['info']['version']
			url = req['releases'][versionPy][0]['url']
			sha256 = req['releases'][versionPy][0]['digests']['sha256']
			if not (url.endswith('.tar.gz') or url.endswith('.zip')):
				try:
					url = req['releases'][versionPy][1]['url']
					sha256 = req['releases'][versionPy][1]['digests']['sha256']
				except:
					url = "error"
					sha256 = "error"
			f.write(oldFormula.format(pyhosted=url,sha256=sha256,version=versionPy))
			f.close()
			if ('-d','') in arguments:
				input('waithing for user input...')
			if 'github' in data['git']:
				uname = data['git']['github'].split('/')[0]
			else:
				uname = input('Your GitHub username: ')
			print('\u001b[4m\u001b[1;36mHomebrew Release:\u001b[0m')
			if 'brew' not in data['git'] and len(os.popen(f'git ls-remote https://github.com/{uname}/homebrew-taps').read())<4:
				print('Creating new repo')
				print('To create a homebrew formula, you must first setup a GitHub repository called "homebrew-taps".\n\nPlease go to https://github.com/new and create a repo called `homebrew-taps`.')
				input('Press enter when done.')
				os.mkdir('homebrew-taps')
				os.chdir('homebrew-taps')
				os.system(f'git init;echo "# homebrew-taps" >> README.md;git remote add origin https://github.com/{uname}/homebrew-taps.git')
				os.chdir('..')
			else:
				print('Cloning existing repo')
				os.system(f'git clone https://github.com/{uname}/homebrew-taps.git')
			if ('-d','') in arguments:
				input('waithing for user input...')
			shutil.copy('dist'+os.sep+'homebrew'+os.sep+data['name']+'.rb','homebrew-taps'+os.sep+data['name']+'.rb')
			os.system(f'cd homebrew-taps;git add .;git config --global credential.helper "cache --timeout=3600";git config user.name "{data["author"]}";git config user.email "{data["email"]}";git commit -F ..{os.sep}.release-notes-latest;git commit --amend -F ..{os.sep}.release-notes-latest;git tag v{version};git push origin master --tags;cd ..')
			data['git']['brew'] = uname+"/homebrew-taps"
		try:
			shutil.rmtree('homebrew-taps')
		except:
			pass
	if "github" in ids or len(ids)==1:
		if 'github' not in data["git"]:
			import sailboat.git
			sailboat.git.main()
		if data['build']['actions_built_latest']:
			up = 'git push origin master;'
		else:
			up = ''
		f = f'git add .;git config --global credential.helper "cache --timeout=3600";git config user.name "{data["author"]}";git config user.email "{data["email"]}";git commit -F .release-notes-latest;git commit --amend -F .release-notes-latest;git tag v{version};git remote add origin https://github.com/{data["git"]["github"]}.git || echo;echo "\u001b[4m\u001b[1;36mGitHub Credentials: (will be cached for 1hr)\u001b[0m";{up}git push origin master --tags;echo "--- done! ---"'
		if up!="":
			print('\n\nPushed twice to update workflow action before tagged push.')
		os.system(f)
		data['build']['actions_built_latest'] = False
		f = open('.'+os.sep+'sailboat.toml','w+')
		f.write(toml.dumps(data))
		f.close()

	print('\n\n\u001b[4m\u001b[1;36mRelease Overview:\u001b[0m')
	print('\nSuccesfully release version \u001b[34m'+version+'\u001b[0m!')
	print('\nView the latest stable releases at:')
	print('PyPi:   				https://pypi.org/project/'+data['short_name'])
	print('GitHub: 				https://github.com/'+data['git']['github'])
	print('GitHub(brew):	https://github.com/'+data['git']['brew'])
	
	


import os
import glob
import sys


class Plugin:
	"""
	The basic sailboat Plugin class
	"""
	name = ""
	data = {}
	description = "<no description provided>"
	version = ""
	options = []
	setup = {}
	_type = "build"
	_show = True
	_release = False
	_order = 0
	_os = 'linux'
	autocompletion = {}

	def registerAutocompletion(self):
		import __main__
		prefix = self.prefix
		f = open(prefix+"autocomplete",'a')
		for x in self.autocompletion.keys():
			f.write(x+"::"+self.autocompletion[x]+"\n")
	def release(self):
		pass
	def init(self):
		pass
	def add(self):
		self.registerAutocompletion()
		print('Installing...')
	def log(self,message):
		print('\u001b[37m[LOG]: '+message+'\u001b[0m')
	def warn(self,message):
		print('\u001b[33m[WARNING]: '+message+'\u001b[0m')
	def error(self,message):
		print('\u001b[31m[LOG]: '+message+'\u001b[0m')
	def __repr__(self):
		return "<sailboat plugin: "+self.name+">"
	def __init__(self,data=None,options=None,prefix=__file__,name=None,version=None):
		if options!=None:
			self.options = options
		if version!=None:
			self.version = version
		if data!=None:
			self.data = data
		self.name = name	
		self.prefix=prefix
		self.init()
	def storeData(self,key,value):
		if self._type == "core":
			self.data[key] = value
		else:
			if self.name not in self.data[self._type]:
				self.data[self._type][self.name] = {}
			self.data[self._type][self.name][key] = value
	def red(self,string):
		if sys.platform.startswith('win'):
			return string
		else:
			return "\u001b[31m"+str(string)+"\033[0m"
	def blue(self,string):
		if sys.platform.startswith('win'):
			return string
		else:
			return "\u001b[34m"+str(string)+"\033[0m"
	def section(self,string):
		if sys.platform.startswith('win'):
			return string
		else:
			return "\n\u001b[4m\u001b[1;36m"+str(string)+"\u001b[0m"
	@property
	def data2(self):
		return self.data[self._type][self.name] 
	def getData(self,key=None):
		try:
			if key==None:
				return self.data[self._type][self.name]
			return self.data[self._type][self.name][key]
		except BaseException as e:
			print('Error in {}:\n\t{}'.format(self.name,e))
			return None
	def getResource(self,name,rwmode='r'):
		prefix = os.path.dirname(os.path.abspath(__file__))+os.sep
		return open(prefix+name,rwmode)
	def wizard(self,setup_dict=None):
		if setup_dict == None:
			setup_dict = self.setup
		for key in setup_dict.keys():
			try:
				t = "::".join(key.split('::')[1:])
				key = key.split('::')[0]
			except:
				t = 'str'
			if self._type!="core" and self.name in self.data[self._type] and key in self.data[self._type][self.name]:
				print(self.blue(setup_dict[str(key)+"::"+t])+" "+str(self.data[self._type][self.name][key]))
			elif self._type=="core" and key in self.data:
				print(self.blue(setup_dict[key+"::"+t])+self.data[key])
			elif t=="str":
				self.storeData(key,input("\u001b[34m"+setup_dict[key+"::"+t]+" \u001b[0m"))
			elif t=="int":
				while True:
					try:
						self.storeData(key,int(input("\u001b[34m"+setup_dict[key+"::"+t]+" \u001b[0m")))
						break
					except ValueError:
						print('Please provide an integer.')
			elif t=="bool":
				try:
					self.storeData(key,input("\u001b[34m"+setup_dict[key+"::"+t]+" [y/n] \u001b[0m")[0]=='y')
				except:
					print('error: you did not supply a value!')
					self.storeData(key,input("\u001b[34m"+setup_dict[key+"::"+t]+" [y/n] \u001b[0m")[0]=='y')
			else:
				self.storeData(key,input("\u001b[34m"+setup_dict[key+"::"+t]+" \u001b[0m"))
		return self.data
	def getFiles(self,extension,recursive=True):
		if recursive:
			files = []
			for root,dirs,files in os.walk('.'):
				for filename in files:
					if filename.endswith(extension):
						files.append(root+filename)
			return files
		else:
			return list(glob.glob(extension))
	def run(self,**kwargs):
		print('This plugin has not been set up correctly.\nIf you are the developer, please add a run(self,**kwargs) function to your class.')
		return



import os
import sys

prefix = os.path.dirname(os.path.abspath(__file__))+os.sep

try:
	f = open(prefix+'autocomplete','r')
	data = f.read().split('\n')
	f.close()
except:
	sys.exit()

for line in data.__reversed__():
	if line == "" or line.startswith('#'):
		continue
	stub = line.split('::')[0]
	completes = line.split('::')[1]
	already_typed = " ".join(sys.argv[1].split(' ')[:-1])
	if stub == already_typed:
		print(completes)



# Copyright 2020 Cole Wilson and other contributors
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
import os
import toml

def main():
	try:
		data = toml.loads(open('.'+os.sep+'sailboat.toml').read())
	except toml.decoder.TomlDecodeError as e:
		print('Config error:\n\t'+str(e))
		exit()
	if data['git']=={}:
		if input('Setup git repo? [y/n] ')[0]=='y':
			al = input('What is your GitHub repo? (leave blank if you don\'t have one yet.)\n<user>/<repo_name>: ')
			if al=="":
				print('Goto https://github.com/new and create a repo called "'+data['short_name']+'."')
				input('Press enter when done.')
				user = input('What is your GitHub username? ')
				al = user+"/"+data['short_name']
			os.system(f'git init;git config --global credential.helper "cache --timeout=3600";git config user.name "{data["author"]}";git config user.email "{data["email"]}";git add .;git commit -m ":rocket: Initial Commit!";git remote add origin https://github.com/{al}.git;git push -u origin master;')
			data['git']['github']=al
		else:
			data['git']['github']=''
	else:
		print('Already set up.')
	open('.'+os.sep+'sailboat.toml','w+').write(toml.dumps(data))


# Copyright 2020 Cole Wilson and other contributors
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
import os
import glob
import toml
import sys
import shutil


def main():
	"""
	The setup wizard for sailboat
	"""
	prefix = os.path.dirname(os.path.abspath(__file__))+os.sep+'resources'
	data = {}
	if os.path.isfile('.'+os.sep+'sailboat.toml'):
		try:
			data = toml.loads(open('.'+os.sep+'sailboat.toml').read())
		except toml.decoder.TomlDecodeError as e:
			print('Config error:\n\t'+str(e))
			exit()
	print("I'm going to ask you a few questions about your project to create a config file!\n\nPlease fill in everything to your ability. If you can't provide a value, leave it blank.\n")
	# GENERAL ================================================
	def red(string):
		if sys.platform.startswith('win'):
			return string
		else:
			return "\033[1;31m"+str(string)+"\033[0m"
	def blue(string):
		if sys.platform.startswith('win'):
			return string
		else:
			return "\033[1;34m"+str(string)+"\033[0m"
	def section(string):
		if sys.platform.startswith('win'):
			return string
		else:
			return "\n\u001b[4m\u001b[1;36m"+str(string)+"\u001b[0m"

	# GENERAL ================================================
	print(section('General Configuration:'))
	questions = {
		"name":"Full name of project",
		"short_name":"Short name of project (unique)",
		"author":"Your name(s)",
		"email":"Project email",
		"short_description":"Short description of your project",
		"description":"Full description of your project",
		"url":"Main project URL",
		"keywords":"Keywords seperated with a space",
		'license':"You seem to have a license file in your project, but what type is it?\n\t- "+(f"\n\t- ".join(lt.split('/'))+"\n>>>")
	}

	licen = len(glob.glob('.'+os.sep+'LICENS*'))>0

	lt = "BSD-2-Clause/BSD-3-Clause/Apache-2.0/LGPL-2.0/LGPL-2.1/LGPL-3.0/GPL-2.0/GPL-3.0/CCDL-1.0/GNU LGPL/MIT/Other SPDX License ID"
	licen = len(glob.glob('.'+os.sep+'LICENS*'))>0
	if licen:
		if 'license' not in self.data:
			print('\033[1;34mYou seem to have a license file in your project, but what type is it?')
			for license in lt.split(os.sep+''):
				print(f'\t- {license}')
			self.data['license'] = input(">>>\033[0m ")
		else:
			print('License: {}'.format(self.data['license']))


	for key in questions:
		if key not in data:
			if key=='license':
				if licen:
					print('\033[1;34mYou seem to have a license file in your project, but what type is it?')
					for license in lt.split(os.sep+''):
						print(f'\t- {license}')
					data[key] = input(">>>\033[0m ")
				else:
					data[key] = 'None'
			else:
				data[key] = input(blue(questions[key])+": ")
		else:
			print(blue(questions[key])+": "+str(data[key]))

	# resource
	print(section('Resource Settings:'))
	if 'resources' not in data:
		data['resources']={}
	questions = {
		"icon":"Square .png icon for your project, leave blank if none",
		"data_files":"List of all data files for your project seperated with spaces. * counts as wildcard",
		"modules":"Space seperated list of modules required for your project",
		"file":"Main python file of your project. Leave blank if none"
	}
	for key in questions:
		if key in data['resources']:
			print(blue(questions[key])+': '+str(data['resources'][key]))
		else:
			if key in ('modules','data_files'):
				data['resources'][key] = input(blue(questions[key])+": ").split()
			else:
				data['resources'][key] = input(blue(questions[key])+": ")
	print(red('Be sure to prefix any paths to any resources with \u001b[4m`os.path.dirname(os.path.abspath(__file__))+os.sep`\u001b[0m\033[1;31m to make sure that they use the correct path and not the current directory.'))

# build
	print(section('Build Settings:'))
	if 'build' not in data:
		data['build']={}
	questions = {
		"type":"Please choose one of the following options for your project:\n\t1. My project only deals with text (not graphical).\n\t2. My project is a graphical app.\n>>>",
		"homebrew":"Would you like to distribute a Homebrew app for your project? [y/n]",
		"unix":"Would you like to distribute a Unix executable for your project? [y/n]",
		"mac":"Would you like to distribute a Mac app for your project? [y/n]",
		"windows":"Would you like to distribute a Windows app for your project? [y/n]",
		
	}
	for key in questions:
		if key in data['build']:
			print(blue(questions[key])+': '+str(data['build'][key]))
		else:
			if "[y/n]" in questions[key]:
				data['build'][key] = input(blue(questions[key])+": ")[0]=='y'
			else:
				data['build'][key] = input(blue(questions[key])+": ")
	if data['build']['type']=="1":
		print(red('Be sure to add some user input in your script, or else the window will close too quickly!'))
	if (data['build']['mac'] or data['build']['windows']):
		if 'installer' in data['build']:
			print(blue("Would you like to provide an installer for the Windows app? (recomended) [y/n]: ")+str(data['build']['installer']))
		else:
			data['build']['installer'] = input(blue("Would you like to provide an installer for the Windows app? (recomended) [y/n]: "))[0]=="y"
	if (not data['build']['mac'] and not data['build']['windows']):
		data['build']['installer'] = False
	if not sys.platform.startswith('win') and data['build']['windows']:
		print(red('You can only make a Windows app on a Windows computer.'))
		needswin=True
	else:
		needswin=False
	if sys.platform != 'darwin' and data['build']['mac']:
		print(red('You can only make a Mac app on a Mac.'))
		needsmac=True
	else:
		needsmac=False

	try:
		asd=data['build']['actions']
	except:
		asd=False

	if (needsmac or needswin) or asd:
		if ('actions' in data['build']):
			print(blue('Github Actions can build your app on the cloud for free.\n\033[1;31mWould you like to use that to build your apps? [y/n]: ')+str(data['build']['actions']))
			actions = data['build']['actions']
		else:
			actions=input('\033[1;34mGithub Actions can build your app on the cloud for free.\n\033[1;31mWould you like to use that to build your apps? [y/n]: \033[0m')[0]=='y'
			data['build']['actions'] = actions
	else:
		actions = False
		data['build']['actions'] = False
	if 'commands' in data['build']:
		print(blue('Commands: ')+str(data['build']['commands']))
	else:
		print(section('Terminal Commands:'))
		if input(blue('Does your project have any terminal commands associated with it? [y/n]: '))[0]=='y':
			data['build']['commands'] = {}
			print(red('`runs:` must be follow the rule \u001b[4m`module.function_name`\u001b[0m\033[1;31m, or be blank, which runs the main project file as a script. \n\nIt is recommended to use function wrappers.\n\nAn example would be:\n\t- name: test_command\n\t  runs: file.test\nThis would run the test() function in file.py\n\nPress ctrl+c to stop asking.'))
			while True:
				try:
					n = input(blue('\t- name: '))
					mod = input(blue('\t  runs: '))
					data['build']['commands'][n]=mod
					print()
				except KeyboardInterrupt:
					print('\nDone!\n\n')
					break
		else:
			data['build']['commands'] = {}

	installl = ''
	if data['build']['mac'] or data['build']['mac'] or data['build']['actions']:
		installl = "### " + " and ".join(["Mac" if data['build']['mac'] else "","Windows" if data['build']['mac'] else "","Unix" if data['build']['actions'] else ""])+f"\nGet the executable from the [latest release](/releases/latest)."
		runn = " or run the application file."
	else:
		runn = "."


	if os.path.isfile('README.md'):
		print(section('README:'))
		if input('\033[1;34mOverwrite current README.md? [y/n]:\033[0m ')[0]=='y':
			open('README.md','w+').write(open(prefix+os.sep+'readme.template.md').read().format(**data,other_install=installl,other_run=runn))
	else:
		open('README.md','w+').write(open(prefix+os.sep+'readme.template.md').read().format(**data,other_install=installl,other_run=runn))

	# Other
	if 'bundle_id' not in data['build']:
		data['build']['bundle_id'] = f"com.{data['author'].lower().replace(' ','')}.{data['short_name']}"
	if 'homebrew' not in data['build']:
		data['build']['homebrew'] = False
	data['name']=data['name'].replace(' ','_').replace('/','-')

	print(section('Git Settings:'))
	if 'git' not in data:
		data['git'] = {}
	print(blue('Please run `sailboat git` to setup GitHub.'))


	f = open('.'+os.sep+'sailboat.toml','w+')
	f.write(toml.dumps(data))
	f.close()

	print('\n\n\033'+red(section('`sailboat.toml` has been created. Please edit that for further changes and advanced options.\033[0m')))


# Copyright 2020 Cole Wilson and other contributors
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import os
import re
import sys
from sailboat.plugins import Plugin
import glob

licensetext = """# Copyright 2020 Cole Wilson and other contributors
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


"""


class Tasks(Plugin):
	description = "Manage 'TODO:' comments"
	_type = "command"
	_show = True
	def run(self):
		prefix = os.path.dirname(os.path.abspath(__file__))+os.sep+'resources'
		if sys.platform.startswith('win'):
			print('windows not yet supported for tasks. if you want, you can help make this possible at github.com/cole-wilson/sailboat')
			return
		tasks = []
		for root,dirs,files in os.walk('.'):
			for filename in files:
				x = os.sep.join((root,filename))
				if x.endswith('.py'):
					f = open(x)
					content = f.read()
					f.close()
					for a in re.findall(r'^.*(#\s*?[tT][oO][dD][oO]:\s?)(.*?)$',content,re.M):
						tasks.append((a,x))
		if len(tasks)==0:
			print('NO TASKS!')
			return
		while True:
			os.system('clear')
			print('\u001b[36m\u001b[4mTODOS:\u001b[0m'.rjust(6))
			print()
			for count,task in enumerate(tasks):
				print(f"{(str(count+1)+'.').ljust(4)}\t{task[0][1].ljust(20)}\t\t({task[1].ljust(10)})")
			try:
				ch = input('\n\nEnter a task number to view code.\n>>> ')
				try:
					ch = int(ch)-1
					content = open(tasks[ch][1]).read()
					content = content.split('\n')
					for count,line in enumerate(content):
						if tasks[ch][0][0] in line:
							linenum = count+1
							break
					line = line.replace(tasks[ch][0][0]+tasks[ch][0][1],'')#.replace('\t','')
					if len(self.options)==2:
						mini = -1*int(self.options[0]) if -1*int(self.options[0]) >1 else -1
						maxi = int(self.options[1]) if int(self.options[1]) >1 else 2
					else:
						mini,maxi = -6,6
					tabsmax = []
					for x in range(mini,maxi):
						tabsmax.append(content[count-x].count('\t'))
					tab = "\t"
					print(f"\n\n\n{linenum+mini}\u001b[32m ... \u001b[31m{content[count+mini].replace(tab,'',min(tabsmax))}\u001b[0m")
					for x in range(mini+1,0):
						print(f"{linenum+x}\u001b[32m ... \u001b[31m{content[count+x].replace(tab,'',min(tabsmax))}\u001b[0m")
					print(f"\u001b[33m{linenum}\u001b[32m ... \u001b[31m{line.replace(tab,'',min(tabsmax))}\u001b[34m\u001b[4m{tasks[ch][0][0]}{tasks[ch][0][1]}\u001b[0m")
					for x in range(1,maxi-1):
						print(f"{linenum+x}\u001b[32m ... \u001b[31m{content[count+x].replace(tab,'',min(tabsmax))}\u001b[0m")
					input(f"{linenum+maxi-1}\u001b[32m ... \u001b[31m{content[count+maxi-1].replace(tab,'',min(tabsmax))}\u001b[0m\n\n\n(press enter)")
				except KeyboardInterrupt:
					print()
					return
				except:
					print('')
			except:
				print('')
				break	


# Copyright 2020 Cole Wilson and other contributors
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

def main():
	import sys
	import getopt
	import os

	opath = os.getcwd()

	while not os.path.isfile('sailboat.toml') and os.getcwd().count(os.sep)>2:
		os.chdir('..')

	if not os.path.isfile('sailboat.toml'):
		print("Couldn't find existing sailboat.toml file")
		os.chdir(opath)
	elif opath != os.getcwd():
		print('Moved to directory `{}`'.format(os.getcwd()))
	he = "Provide a mode:\n\t"+sys.argv[0].split('/')[-1]+" [wizard | build | dev | git | release | task]"

	try:
		arguments, ids = getopt.getopt(sys.argv[1:], "dfyn", ['pypi-only','pyinstaller-only','homebrew-only','actions-only','windows-only','mac-only'])
	except:
		print(he)
		sys.exit(0)

	noint = ('-y','') in arguments

	try:
		mode = ids[0]
	except:
		print(he)
		sys.exit(0)


	# ================================================================

	if mode == "wizard":
		import sailboat.wizard as wizard
		wizard.main()
	elif mode.startswith('task'):
		import sailboat.task as task
		task.main(ids)

	elif mode == "build":
		import sailboat.build as build
		build.main(ids,arguments,nointeraction=noint)

	elif mode == "release":
		import sailboat.release as release
		release.main(arguments,ids)
	# elif mode in ["build","dev","release"]:

	elif mode == "dev":
		os.system('python3 ./setup.py develop')

	elif mode == "git":
		import sailboat.git as git
		git.main()
	else:
		print('Invalid option')
		print(he)

	# 	elif mode == "dev":
	# 		os.system('python3 .'+os.sep+'setup.py develop')
	# 	for x in glob.glob('*.egg-info'):
	# 		shutil.rmtree(x)
	# else:
	# 	print(f'Illegeal option `{mode}`')
	# 	sys.exit(0)

	# if mode=="release":
	# 	print("Please make sure that you have a https://pypi.org/ account.")
	# 	try:
	# 		import twine
	# 	except:
	# 		input('Press enter to continue installing `twine`. Press ctrl+x to exit.')
# 		os.system('python3 -m pip install --user --upgrade twine || python3 -m pip install --upgrade twine')
# 	os.system('python3 -m twine release dist'+os.sep+'*')
if __name__=='__main__':
	main()


import sys
import os
import pkg_resources
import toml
import json
import sailboat.plugins
import sailboat.core
from sailboat.build import Build
from pathlib import Path

try:
	path = Path(__file__)
	prefix = os.path.abspath(path.parent)+os.sep
except:
	prefix = os.path.dirname(os.path.abspath(__file__))+os.sep

__doc__ = "usage: sail [options ...] [command]\n"

def refreshEntries():
	plugins={"core":{},"build":{},"release":{},"command":{}}
	for entry_point in pkg_resources.iter_entry_points(group='sailboat_plugins'):
		temp = entry_point.load()
		plugins[temp._type][entry_point.name] = {
			"show" : temp._show,
			"dist" : str(entry_point.dist).split(' ')[0],
			"description" : temp.description,
			"type" : temp._type,
			'release' : temp._release,
			"order" : temp._order,
			"default_os":str(temp._os)
		}
	f = open(os.path.abspath(path.parent)+os.sep+'plugins.json','w+')
	f.write(json.dumps(plugins, indent=2, sort_keys=True))
	f.close()
	del f
	return plugins

def main():
	global __doc__
	opath = os.getcwd()
	while not os.path.isfile('sailboat.toml') and os.getcwd().count(os.sep)>2:
		os.chdir('..')
	if not os.path.isfile('sailboat.toml'):
		needswizard = True
		data = {}
		os.chdir(opath)
	else:
		needswizard = False
		with open('sailboat.toml') as file:
			data = toml.loads(file.read())
			if data == {}:
				needswizard = True
	needswizard =False
	if 'command' not in data:
		data['command'] = {}
	if 'build' not in data:
		data['build'] = {}
	if 'release' not in data:
		data['release'] = {}
	if 'git' not in data:
		data['git'] = {}
	
	switches = []
	startindex = 0
	for index,value in enumerate(sys.argv[1:]):
		if value.startswith('--'):
			switches.append(value[2:])
		elif value.startswith('-'):
			switches.extend(value[1:])
		else:
			startindex = index
			break		
	options = [sys.argv[0],*sys.argv[1:][startindex:]]

# =============================================================================
	plugins = json.loads(open(prefix+'plugins.json').read())
	
	if 'refresh' in switches or 'r' in switches or plugins=={} or (len(plugins.keys())!=4):
		plugins = refreshEntries()
		print('reloaded plugins')

	a_commands = []

	if len(plugins['core'].keys())!=0:
		__doc__+="\n\tcore commands:"

	for command in plugins['core'].keys():
		if plugins['core'][command]['show']:
			__doc__+="\n\t\t- \u001b[36;1m"+command+"\u001b[0m: "+plugins['core'][command]['description'].capitalize()
			a_commands.append(command)

	__doc__+="\n\t\t- \u001b[36;1mhelp\u001b[0m: Display this message."

	if len(plugins['command'].keys())!=0:
		__doc__+="\n\n\tother commands:"

	for command in plugins['command'].keys():
		if plugins['command'][command]['show'] and command in data['command']:
			__doc__+="\n\t\t- \u001b[36;1m"+command+"\u001b[0m: "+plugins['command'][command]['description'].capitalize()
			a_commands.append(command)

	__doc__+="\n"

# =============================================================================

	if len(options) < 2:
		print(__doc__)
		return
	if needswizard:
		command = 'wizard'
	else:
		command = options[1]
	autocomplete = False
	if command == 'help' or 'help' in switches or 'h' in switches:
		print(__doc__)
		return

	if command in plugins['core']:
		t = 'core'
	elif command in plugins['command']:
		t = 'command'
	else:
		if not autocomplete:
			print('sailboat: error: {} is not a valid command. Please make sure you have installed it.'.format(command))
		return	

	if t!='core' and command not in data['build'] and command not in data['release'] and command not in data['command']:
		print('sailboat: error: {} *is* a valid command, but it isn\'t installed on this project. Install it with the `add` command.'.format(command))
		return

	dist = plugins[t][command]['dist']
	temp = pkg_resources.load_entry_point(dist,'sailboat_plugins',command)
	temp = temp(
		data=data,
		options=options[2:],
		name=command,
		prefix=prefix
	)
	if autocomplete:
		print(" ".join(temp.autocomplete()))
	elif t == "core":
		temp.run(plugins=plugins)
	else:
		temp.run()

	if t == 'core':
		data = temp.data
	else:
		data[t][command] = temp.data[t][command]

	basic_data = {}
	resources = {'resources':{}}
	commands = {'command':{}}
	builds = {'build':{}}
	release = {'release':{}}
	other = {}

	with open('sailboat.toml','w+') as f:
		for key in data.keys():
			if not isinstance(data[key],dict):
				basic_data[key] = data[key]
			elif key == 'resources' and len(data[key].values())>0:
				resources['resources'] = data[key]
			elif key == 'command' and len(data[key].values())>0:
				commands['command'] = data[key]
			elif key == 'build' and len(data[key].values())>0:
				builds['build'] = data[key]
			elif key == 'release' and len(data[key].values())>0:
				release['release'] = data[key]
			elif key not in ('build','release','command','resources'):
				other[key] = data[key]
		resources  = resources if resources != {'resources':{}} else {}
		commands  = commands if commands != {'command':{}} else {}
		builds  = builds if builds != {'build':{}} else {}
		release  = release if release != {'release':{}} else {}
		other  = other if other != {'other':{}} else {}
		o = [*map(toml.dumps,[basic_data,resources,commands,builds,release,other])]
		out="""#	  _____       _ _ _                 _     
#	 / ____|     (_) | |               | |  _ 
#	| (___   __ _ _| | |__   ___   __ _| |_(_)
#	 \___ \ / _` | | | '_ \ / _ \ / _` | __|  
#	 ____) | (_| | | | |_) | (_) | (_| | |_ _ 
#	|_____/ \__,_|_|_|_.__/ \___/ \__,_|\__(_)
                                          
# Basic Setup:
{}

# Resource Setup:
{}

# Plugin Commands:
{}

# Build Routines:
{}

# Release Routines:
{}

# Other:
{}

# Thank you for using Sailboat!"""
		if '_comments' in data and not data['_comments']:
			print()
			out = "{}\n{}\n{}\n{}\n{}\n{}\n"
		out = out.format(o[0],o[1],o[2],o[3],o[4],o[5])
		f.write(out)

if __name__ == "__main__":
	main()


# Copyright 2020 Cole Wilson and other contributors
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
import os
import sys
import glob
import shutil
import requests
import time
from sailboat.plugins import Plugin

class PyPi(Plugin):
	_type = "build"
	_release = True
	def release(self):
		print("Please make sure that you have a https://pypi.org/ account.")
		try:
			import twine
		except:
			input('Press enter to continue installing `twine`. Press ctrl+c to exit.')
			os.system('python3 -m pip install --user --upgrade twine || python3 -m pip install --upgrade twine')
			import twine
		del twine
		print('\u001b[4m\u001b[1;36mPyPi Credentials:\u001b[0m')
		os.system('python3 -m twine upload dist'+os.sep+'pypi'+os.sep+'*')

	def wizard(self):
		print(self.section('Terminal Commands:'))
		if 'commands' in self.data[self._type][self.name]:
			print(self.blue('Commands: ')+str(self.data[self._type][self.name]['commands']))
		elif input(self.blue('Does your project have any terminal commands associated with it? [y/n]: '))[0]=='y':
			self.storeData('commands',{})
			print(self.red('`runs:` must be follow the rule \u001b[4m`module.function_name`\u001b[0m\033[1;31m, or be blank, which runs the main project file as a script. \n\nIt is recommended to use function wrappers.\n\nAn example would be:\n\t- name: test_command\n\t  runs: file.test\nThis would run the test() function in file.py\n\nPress ctrl+c to stop asking.'))
			while True:
				try:
					n = input(self.blue('\t- name: '))
					mod = input(self.blue('\t  runs: '))
					self.data[self._type][self.name]['commands'][n]=mod
					print()
				except KeyboardInterrupt:
					print('\nDone!\n\n')
					break
		else:
			self.data[self._type][self.name]['commands'] = {}

		print(self.section('Entry points:'))
		if 'entry_points' in self.data[self._type][self.name]:
			print(self.blue('Entry points: ')+str(self.data[self._type][self.name]['entry_points']))
		elif input(self.blue('Does your project have any entry points associated with it? [y/n]: '))[0]=='y':
			self.storeData('entry_points',{})
			stillGo = True
			while True:
				try:
					group = input('\t- Group name: ')
					if group not in self.data[self._type][self.name]['entry_points']:
						self.data[self._type][self.name]['entry_points'][group] = {}
					while True:
						try:
							name = input('\t\t- Entrypoint Name: ')
							locator = input('\t\t  Locator: ')
							self.data[self._type][self.name]['entry_points'][group][name] = locator
							print()
						except KeyboardInterrupt:
							print('\n')
							break
				except KeyboardInterrupt:
					print('\n')
					break
		else:
			self.storeData('entry_points',{})
	def run(self,**kwargs):
	# ============== Create bin Script ===============================================
		try:
			os.mkdir('bin')
		except FileExistsError:
			pass
		bins = []
		for commandname in self.getData('commands').keys():
			if self.getData('commands')[commandname] == '':
				open("bin"+os.sep+commandname,'w+').write(f"#!"+os.sep+"usr"+os.sep+"bin"+os.sep+f"env bash\npython3 -m {self.data['short_name']} $@")
				bins.append('bin'+os.sep+commandname)
		# ============== Generate setup.py ===============================================
		print('\n\n\u001b[4m\u001b[1;36mGenerating setup.py\u001b[0m')
		if 'custom_setup' in self.data:
			cu=str(self.data['custom_setup'])
		else:
			cu=str({})
		with self.getResource('resources/setup.py.template') as datafile:
			template = datafile.read()

		entries = {'console_scripts':[]}
		for commandname in self.getData('commands').keys():
			if self.getData('commands')[commandname]!="":
				modname = ".".join(self.getData('commands')[commandname].split('.')[:-1])
				funcname = self.getData('commands')[commandname].split('.')[-1]
				entries['console_scripts'].append(commandname+"="+self.data["short_name"]+"."+modname+":"+funcname)

		for entry_point_group in self.getData('entry_points').keys():
			if entry_point_group not in entries:
				entries[entry_point_group] = []
			for entry_point_name in self.getData('entry_points')[entry_point_group]:
				entries[entry_point_group].append(entry_point_name+"="+self.getData('entry_points')[entry_point_group][entry_point_name])

		try:
			pyv = self.version.split('+')[0]
		except:
			pyv = self.version

		setup = template.format(
			**self.data,
			**self.data['resources'],
			cu=cu,
			bins=bins,
			version = pyv,
			entry_points = entries
		)
		open('setup.py','w+').write(setup)
	
		# ============== Generate pypi files ===============================================
		print('\n\n\u001b[4m\u001b[1;36mGenerating PyPi files...\u001b[0m')
		os.system('python3 .'+os.sep+'setup.py bdist_wheel sdist -d dist'+os.sep+'pypi')
		try:
			shutil.rmtree('build')
		except:
			pass
		for x in glob.glob('dist'+os.sep+'*.whl'):
			os.rename(x,x.replace('dist'+os.sep,'dist'+os.sep+'pypi'+os.sep))
		for x in glob.glob('*.egg-info'):
			shutil.rmtree(x)

class Homebrew(Plugin):
	_type = "build"
	_release = True
	_order = 200

	def release(self):
		self.options = self.options[0]
		if "pypi" in self.options or len(self.options)==0:
			print('Must wait 20 seconds for pypi files to upload before creating Homebrew formula...\n')
			for i in range(0, 20):
				time.sleep(1)
				sys.stdout.write(u"\u001b[1000D" + str(20-(i + 1)).zfill(2) + " seconds left.\t")
				sys.stdout.flush()
			print()
		try:
			shutil.rmtree('homebrew-taps')
		except:
			pass
		f = open('dist'+os.sep+'homebrew'+os.sep+self.data['name']+'.rb')
		oldFormula = f.read()
		f.close()
		f = open('dist'+os.sep+'homebrew'+os.sep+self.data['name']+'.rb','w+')
		req = requests.get('https://pypi.org/pypi/{}/json'.format(self.data['short_name'])).json()
		versionPy = req['info']['version']
		url = req['releases'][versionPy][0]['url']
		sha256 = req['releases'][versionPy][0]['digests']['sha256']
		if not (url.endswith('.tar.gz') or url.endswith('.zip')):
			try:
				url = req['releases'][versionPy][1]['url']
				sha256 = req['releases'][versionPy][1]['digests']['sha256']
			except:
				url = "error"
				sha256 = "error"
		f.write(oldFormula.format(pyhosted=url,sha256=sha256,version=versionPy))
		f.close()
		if 'github' in self.data['git']:
			uname = self.data['git']['github'].split('/')[0]
		else:
			uname = input('Your GitHub username: ')
		print('\u001b[4m\u001b[1;36mHomebrew Release:\u001b[0m')
		if 'brew' not in self.data['git'] and len(os.popen(f'git ls-remote https://github.com/{uname}/homebrew-taps').read())<4:
			print('Creating new repo')
			print('To create a homebrew formula, you must first setup a GitHub repository called "homebrew-taps".\n\nPlease go to https://github.com/new and create a repo called `homebrew-taps`.')
			input('Press enter when done.')
			os.mkdir('homebrew-taps')
			os.chdir('homebrew-taps')
			os.system(f'git init;echo "# homebrew-taps" >> README.md;git remote add origin https://github.com/{uname}/homebrew-taps.git')
			os.chdir('..')
		else:
			print('Cloning existing repo')
			os.system(f'git clone https://github.com/{uname}/homebrew-taps.git')
		shutil.copy('dist'+os.sep+'homebrew'+os.sep+self.data['name']+'.rb','homebrew-taps'+os.sep+self.data['name']+'.rb')
		# input()
		os.system(f'cd homebrew-taps;git add .;git config --global credential.helper "cache --timeout=3600";git config user.name "{self.data["author"]}";git config user.email "{self.data["email"]}";git commit -m "{self.data["release-notes"]}";git commit --amend -m "{self.data["release-notes"]}";echo git tag v{self.version};git push origin master --tags;cd ..')
		self.data['git']['brew'] = uname+"/homebrew-taps"
		try:
			shutil.rmtree('homebrew-taps')
		except:
			pass

	def run(self,**kwargs):
		retmp = '   resource "{name}" do\n      url "{url}"\n      sha256 "{sha256}"\n   end\n'
		resources = ''
		for modulename in self.data['resources']['modules']:
			req = requests.get('https://pypi.org/pypi/{}/json'.format(modulename)).json()
			versionPy = req['info']['version']
			url = req['releases'][versionPy][0]['url']
			sha256 = req['releases'][versionPy][0]['digests']['sha256']
			if not (url.endswith('.tar.gz') or url.endswith('.zip')):
				try:
					url = req['releases'][versionPy][1]['url']
					sha256 = req['releases'][versionPy][1]['digests']['sha256']
				except:
					continue
			resources+=retmp.format(name=modulename,url=url,sha256=sha256)
		os.makedirs('dist'+os.sep+'homebrew')
		f = open('dist'+os.sep+'homebrew'+os.sep+'{name}.rb'.format(name=self.data['name']),'w+')
		f.write(self.getResource('resources'+os.sep+"brew.rb").read().format(
			**self.data,
			resources2 = resources,
			version = self.version
		))
		f.close()

class PyInstaller(Plugin):
	_type = "build"
	setup = {
		"type::int":"Select one of the following:\n\t1. My app needs a console window to display input and output.\n\t2. My app should provide it's own windows.\n>>> ",
		"mac::bool":"Would you like to generate a Mac app for your project? ",
		"windows::bool":"Would you like to generate a Windows app for your project? ",
		"unix::bool":"Would you like to generate a Unix executable for your project? ",
	}
	_os = "windows mac linux"

	def wizard(self):
		super().wizard()
		if 'bundle_id' not in self.getData():
			self.storeData('bundle_id',"com."+self.data['author'].lower().replace(' ','')+"."+self.data['short_name'])

	def run(self,**kwargs):
		if not(self.getData('mac') or self.getData('windows') or self.getData('unix')):
			return
		try:
			import PyInstaller.__main__
		except:
			print('Installing PyInstaller...')
			os.system('pip3 install pyinstaller')
			import PyInstaller.__main__
		
		mods = []
		for x in self.data['resources']['modules']:
			mods.append('--hidden-import')
			mods.append(x)

		datafiles = []
		for x in self.data['resources']['data_files']:
			for g in glob.glob(self.data['short_name']+os.sep+x):
				datafiles.append('--add-data')
				datafiles.append(g+os.pathsep+g.replace(self.data['short_name']+os.sep,''))
		typ = '--nowindowed' if self.getData('type')==1 else '--noconsole'

		ico = ['--icon',self.data['resources']['icon']] if 'icon' in self.data['resources'] else []
		options = [
			self.data['short_name']+os.sep+'__main__.py',
			'--onefile',
			'--name',
			self.data['name'],
			'--distpath',
			'.'+os.sep+'dist'+os.sep+'pyinstaller',
			*mods,
			*datafiles,
			typ,
			*ico,
			'--osx-bundle-identifier',
			self.getData('bundle_id')
		]
		print(options)
		PyInstaller.__main__.run(options)
		try:
			print('removing '+self.data['name']+".spec...")
			os.remove(self.data['name']+".spec")
		except:
			print('removing app.spec...')
			os.remove("app.spec")			
    # MAC APP BUNDLE==============
		print('\n\n\u001b[4m\u001b[1;36mGenerating Mac .app bundle...\u001b[0m')
		if self.getData('mac') and sys.platform.startswith('dar'):
			os.chdir('dist')

			os.mkdir(self.data['name'])
			os.chdir(self.data['name'])

			os.mkdir('Contents')
			os.chdir('Contents')

			os.mkdir('MacOS')
			os.mkdir('Resources')

			infoPlist = open('Info.plist','w+')
			infoPlist.write(self.getResource('resources'+os.sep+'info.plist.xml').read().format(
				**self.data,
				**self.data['build'],
				version = self.version,
				bundle_id = self.getData('bundle_id')
			))
			infoPlist.close()

			shutil.copy('./../../pyinstaller/'+self.data['name'],'MacOS')
			os.chdir('./../../..')

			os.rename('./dist/'+self.data['name'],'./dist/'+self.data['name']+".app")
		else:
			print('not generating mac .app bundle because on {} not mac.'.format(sys.platform))
		# ============== Generate Installer Package ===============================================
		print('\n\n\u001b[4m\u001b[1;36mGenerating Installer Package...\u001b[0m')
		if not (self.getData('windows') or self.getData('unix') or self.getData('mac')):
			pass
		elif sys.platform.startswith('darwin'):#MAC
			try:
				import dmgbuild
			except:
				print('Installing dmgbuild')
				os.system('pip3 install dmgbuild')
				import dmgbuild
			del dmgbuild
			if not os.path.isdir('build'):
				os.mkdir('build')
			open('build'+os.sep+'settings.py','w+').write(self.getResource('resources'+os.sep+'settings.py.template').read().format(
				**self.data,
				version=self.version,
				icns=self.data['resources']['icon'],
				keywo=", ".join(self.data['keywords'])
			))
			os.system(f'cat build/settings.py;dmgbuild -s .{os.sep}build{os.sep}settings.py "{self.data["name"]} Installer" ./{self.data["name"]}.dmg')

		else:
			print(f'Installer creation not yet supported for {sys.platform}!')




#!/usr/bin/env python3

# +------------------------+			   
# | Created with Sailboat  |			   
# |			                   |			   
# | Do not edit this file  |			   
# | directly. Instead  	   |			   
# | you should edit the	   |			   
# | `sailboat.toml` file.  |			   
# +------------------------+	

import setuptools

try:
  with open("README.md", "r") as fh:
	  long_description = fh.read()
except:
	long_description = "# Sailboat\n🐍 A quick and easy way to distribute your Python projects!\n### Contributors\n- Cole Wilson\n### Contact\n<cole@colewilson.xyz> "

options = {
	"name":"sailboat",
	"version":"0.19.0",
	"scripts":[],
	"entry_points":{'console_scripts': ['sail=sailboat.__main__:main', 'sailboat=sailboat.__main__:main'], 'sailboat_plugins': ['quickstart=sailboat.core:QuickStart', 'pypi=sailboat.builders:PyPi', 'homebrew=sailboat.builders:Homebrew', 'pyinstaller=sailboat.builders:PyInstaller', 'dev=sailboat.core:Develop', 'release=sailboat.core:Release', 'build=sailboat.build:Build', 'wizard=sailboat.core:Wizard', 'remove=sailboat.core:Remove', 'add=sailboat.core:Add', 'plugins=sailboat.core:ManagePlugins', 'git=sailboat.core:Git', 'tasks=sailboat.other:Tasks', 'workflow=sailboat.core:Actions', 'github=sailboat.core:Git', 'github_release=sailboat.core:GithubRelease']},
	"author":"Cole Wilson",
	"author_email":"cole@colewilson.xyz",
	"description":"🐍 A quick and easy way to distribute your Python projects!",
	"long_description":long_description,
	"long_description_content_type":"text/markdown",
	"url":"https://github.com/cole-wilson/sailboat",
	"packages":setuptools.find_packages(),
	"install_requires":['toml', 'semver', 'requests', 'setuptools'],
	"classifiers":["Programming Language :: Python :: 3"],
	"python_requires":'>=3.6',
	"package_data":{"": ['resources/*', 'plugins.json', 'autocomplete', 'auto.bash'],},
	"license":"MIT",
	"keywords":'',
}

custom_options = {}

if __name__=="__main__":
	setuptools.setup(**custom_options,**options)